/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with {library}; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.commons.database;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.persistence.PersistenceException;
import javax.sql.DataSource;

import org.hibernate.ejb.EntityManagerFactoryImpl;
import org.hibernate.stat.QueryStatistics;
import org.hibernate.stat.Statistics;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.sonar.commons.database.dialect.JdbcDialect;

public class DefaultDatabaseConnector implements DatabaseConnector {

  private static final Logger LOG = LoggerFactory.getLogger(DefaultDatabaseConnector.class);
  private static final Logger LOG_SQL = LoggerFactory.getLogger("org.hibernate.SQL");
  private static final Logger LOG_STATISTICS = LoggerFactory.getLogger("org.sonar.DBSTATISTICS");

  /**
   * System property name to force the dialect class name (com.foo.bar.MyHibernateDialectClassName) to use.
   * Can be useful when the dialect impl autodetected does not work correctly with your database.
   */
  public final static String HIBERNATE_JDBC_DIALECT_IMPL = "org.codehaus.sonar.jdbc.dialect.impl";
  /**
   * System property name to force the dialect name (one value of {@link ch.hortis.sonar.model.JdbcData.JDBCDialect}) to use.
   * Can be useful when the dialect name cannot be auto detected from the provided JDBC connection URL
   */
  public final static String HIBERNATE_JDBC_DIALECT_NAME = "org.codehaus.sonar.jdbc.dialect.name";


  private final DatabaseConfiguration configuration;
  private EntityManagerFactory factory = null;
  private int databaseVersion = -1;
  private boolean started=false;

  public DefaultDatabaseConnector(DatabaseConfiguration configuration) {
    this.configuration = configuration;
  }

  public DatabaseConfiguration getConfiguration() {
    return configuration;
  }

  /**
   * throws PersistenceException if schema is not valid
   */
  public void start() throws WrongDatabaseVersionException {
    if (configuration.isCreateDatabaseFromHibernateMapping()) {
      // hibernate creates the schema
      getEntityManagerFactory();
      populateDatabase();
      checkVersion();

    } else {
      checkVersion();

      // the HibernateFactory validates the database
      getEntityManagerFactory();
    }
    started=true;
  }

  public void stop() {
    if (factory != null && factory.isOpen()) {
      factory.close();
      factory = null;
      started=false;
    }
  }

  private int loadVersion() {
    Connection connection = null;
    Statement stmt = null;
    ResultSet rs = null;
    ResultSet tablesRs = null;
    try {
      connection = getConnection();
      tablesRs= connection.getMetaData().getTables(null, null, "%", null);
      boolean tableExists=false;
      while(tablesRs.next()) {
        // the 3rd column is the table name :
        // http://java.sun.com/j2se/1.4.2/docs/api/java/sql/DatabaseMetaData.html#getTables(java.lang.String,%20java.lang.String,%20java.lang.String,%20java.lang.String[])
        if ("SCHEMA_INFO".equalsIgnoreCase(tablesRs.getString(3))) {
          tableExists=true;
          break;
        }
      }
      if (tableExists) {
        stmt = connection.createStatement();
        rs = stmt.executeQuery("SELECT version FROM " + SchemaInfo.TABLE_NAME);
        if (rs.next()) {
          return rs.getInt(1);
        }
      }

    } catch (SQLException e) {
      throw new PersistenceException("Can not check the database version", e);

    } finally {
      if (rs != null) {
        try {
          rs.close();
        } catch (SQLException e) {

        }
      }
      if (tablesRs != null) {
        try {
          tablesRs.close();
        } catch (SQLException e) {

        }
      }
      if (stmt != null) {
        try {
          stmt.close();
        } catch (SQLException e) {
        }
      }
      try {
        if (connection != null) {
          connection.close();
        }
      } catch (SQLException e) {

      }
    }

    return -1;
  }

  private boolean isValidVersion() {
    if (databaseVersion == DatabaseConfiguration.DATABASE_VERSION) {
      return true;
    }
    databaseVersion = loadVersion();
    return (databaseVersion == DatabaseConfiguration.DATABASE_VERSION);
  }

  public void checkVersion() throws WrongDatabaseVersionException {
    if (!isValidVersion()) {
      throw new WrongDatabaseVersionException(databaseVersion, DatabaseConfiguration.DATABASE_VERSION);
    }
  }


  public Connection getConnection() throws SQLException {
    Connection jdbcConn;
    if (configuration.getDatasourceName() != null) {
      DataSource ds;
      try {
        Context ctx = new InitialContext();
        ds = (DataSource) ctx.lookup(configuration.getDatasourceName());

      } catch (Exception ex) {
        SQLException sqlEx = new SQLException("Error during datasource JNDI lookup");
        sqlEx.initCause(ex);
        throw sqlEx;
      }
      jdbcConn = ds.getConnection();

    } else {
      try {
        Class.forName(configuration.getDriverClassName());

      } catch (ClassNotFoundException e) {
        throw new SQLException("Cannot found sql driver " + configuration.getDriverClassName());
      }
      jdbcConn = DriverManager.getConnection(configuration.getUrl(), configuration.getUsername(), configuration.getPassword());
    }
    return jdbcConn;
  }

  protected final void populateDatabase() {
    SchemaInfo schemaInfo = new SchemaInfo();
    schemaInfo.setVersion(configuration.getCreateDatabaseVersion());
    EntityManager manager = null;
    try {
      manager = createEntityManager();
      manager.getTransaction().begin();
      manager.persist(schemaInfo);
      manager.getTransaction().commit();

      LOG.info("Database created");

    } finally {
      if (manager != null) {
        manager.close();
      }
    }
  }


  public EntityManager createEntityManager() throws WrongDatabaseVersionException {
    EntityManager manager = getEntityManagerFactory().createEntityManager();
    if (LOG_STATISTICS.isInfoEnabled()) {
      manager = new StatisticsEntityManager(manager, LOG_STATISTICS);
    }
    return manager;
  }


  public synchronized EntityManagerFactory getEntityManagerFactory() {
    if (factory != null) {
      return factory;
    }

    // other settings are stored into /META-INF/persistence.xml
    Properties props = new Properties();
    props.put("hibernate.connection.isolation", Integer.toString(configuration.getIsolationLevel()));
    props.put("hibernate.hbm2ddl.auto", configuration.getDdlMode());


    if (configuration.getDatasourceName() == null) {
      props.put("hibernate.connection.url", configuration.getUrl());
      props.put("hibernate.connection.driver_class", configuration.getDriverClassName());
      props.put("hibernate.connection.username", configuration.getUsername());
      props.put("hibernate.connection.password", configuration.getPassword());
      props.put("hibernate.dialect", JdbcDialect.getHibernateDialect(configuration.getUrl()));

    } else {
      props.put("hibernate.connection.datasource", configuration.getDatasourceName());
      props.put("hibernate.dialect", JdbcDialect.getHibernateDialect(configuration.getJdbcDialect()));
    }

    if (LOG_STATISTICS.isInfoEnabled()) {
      props.put("hibernate.generate_statistics", "true");
    }
    if (LOG_SQL.isInfoEnabled()) {
      props.put( "hibernate.show_sql", "true");
    }

    if (configuration.isCacheEnabled()) {
      props.put("hibernate.cache.provider_class", "org.hibernate.cache.EhCacheProvider");
      props.put("hibernate.cache.use_second_level_cache", "true");
      props.put("hibernate.cache.use_query_cache", "false");

    } else {
      props.put("hibernate.cache.provider_class", "org.hibernate.cache.NoCacheProvider");
      props.put("hibernate.cache.use_second_level_cache", "false");
      props.put("hibernate.cache.use_query_cache", "false");
    }

    factory = Persistence.createEntityManagerFactory("sonar", props);
    return factory;
  }

  public boolean isStarted() {
    return started;
  }

  public void dumpStatistics() {
    if (LOG_STATISTICS.isInfoEnabled() &&
        getEntityManagerFactory() instanceof EntityManagerFactoryImpl) {

      EntityManagerFactoryImpl hibernateFactory = (EntityManagerFactoryImpl) getEntityManagerFactory();
      Statistics stats = hibernateFactory.getSessionFactory().getStatistics();
      for (String query : stats.getQueries()) {
        QueryStatistics stat = stats.getQueryStatistics(query);
        StringBuilder sb = new StringBuilder();
        sb.append(stats.getQueryExecutionMaxTime());
        sb.append("ms, count=");
        sb.append(stat.getExecutionCount());
        sb.append(", sql=");
        sb.append(stat.getCategoryName());
        LOG_STATISTICS.info(sb.toString());
      }
      stats.clear();
    }
  }

}